from typing import List,  Optional

from legion_utils import Priority, priority_of
from pydantic import validator
from pydantic.dataclasses import dataclass


class Strict:
    extra = "forbid"

@dataclass(config=Strict)
class FailurePriorityThreshold:
    num_failures: int
    priority: Priority

    @validator('priority', pre=True, allow_reuse=True)
    def _priority(cls, v: str) -> Priority:  # pylint: disable=E0213,R0201
        return priority_of(v)


@dataclass(config=Strict)
class LegionConfig:
    exchange: str
    route: str
    failure_priority_thresholds: List[FailurePriorityThreshold]
    reporting_delay_seconds: float


@dataclass(config=Strict)
class TaskConfig():
    task_id: str
    delay_seconds: float
    command: str
    legion: LegionConfig
    on_failure_delay_seconds: Optional[float] = None



from functools import partial
from pathlib import Path
from shlex import split
from subprocess import STDOUT, Popen, PIPE
from threading import Event
from typing import List, Optional, Generator

from typeguard import typechecked

from testudo.config import TaskConfig
from testudo.log import log
from testudo.reporter import start_reporter_daemon
from testudo.task_manager import task_manager

@typechecked
def report_failure(db: Path, task_id: str, output: str, exit_code: int, stack_trace: Optional[str] = None) -> None:
    with task_manager(db) as tm:
        tm.report_failure(task_id, output, exit_code, stack_trace)


@typechecked
def report_success(db: Path, task_id: str, output: str) -> None:
    with task_manager(db) as tm:
        tm.report_success(task_id, output)

@typechecked
class OutputWindow:
    def __init__(self, limit: Optional[int] = None) -> None:
        self.limit = limit
        self.contents: List[str] = []

    def append(self, line: str) -> None:
        self.contents.append(line)
        if self.limit is not None:
            while len(self.contents) > self.limit:
                self.contents.pop(0)

    @property
    def output(self) -> str:
        joined = '\n'.join(self.contents)
        return joined

@typechecked
def run(db: Path, task_id: str, cmd: List[str]) -> bool:
    log.info(f"Running Task [{task_id}]...")
    out = OutputWindow(30)
    with Popen(cmd, stderr=STDOUT, stdout=PIPE) as proc:
        line = []
        while proc.poll() is None:
            if proc.stdout is None:  # pragma: no branch
                break  # pragma: no cover
            for byte in iter(partial(proc.stdout.read, 1), b''):
                char = byte.decode()
                if char not in {'\n', '\r', ''}:
                    line.append(char)
                else:
                    joined = "".join(line)
                    log.info(f'[{task_id}] OUTPUT: {joined}')
                    out.append(joined.rstrip())
        exit_code = proc.returncode
        if exit_code == 0:
            log.info(f"Task [{task_id}] successful!")
            report_success(db, task_id, out.output)
            return True
        log.warning(f"Task [{task_id}] failed!")
        report_failure(db, task_id, '\n'.join(out.output), exit_code)
        return False

@typechecked
def run_with_delay(db: Path, task_id: str,
                   cmd: List[str],
                   delay_seconds: float,
                   halt_flag: Event,
                   on_failure_delay_seconds: Optional[float] = None) -> Generator[bool, None, None]:
    on_failure_delay_seconds = on_failure_delay_seconds or delay_seconds
    while 42:
        success = run(db, task_id, cmd)
        yield success
        if halt_flag.wait(delay_seconds if success else on_failure_delay_seconds):
            break

@typechecked
def run_with_reporter(db: Path, config: TaskConfig, halt: Optional[Event] = None) -> None:
    first_run_complete = False
    _halt = halt or Event()
    for _ in run_with_delay(db, config.task_id, split(config.command),
                            delay_seconds=config.delay_seconds, halt_flag=_halt,
                            on_failure_delay_seconds=config.on_failure_delay_seconds):
        if not first_run_complete:
            first_run_complete = True
            start_reporter_daemon(db, config, halt=_halt)



from math import ceil
from pathlib import Path
from pprint import pformat
from socket import gethostname
from threading import Thread, Event
from traceback import format_exc
from typing import Optional

from funcy import first
import legion_utils
from legion_utils import Priority, NotificationMsg
from typeguard import typechecked

from testudo.config import TaskConfig
from testudo.log import log
from testudo.task_manager import task_manager

HOSTNAME = gethostname()


@typechecked
def report(db: Path, task_config: TaskConfig) -> NotificationMsg:
    with task_manager(db) as tm:
        if recent_failures := list(f for f in tm.get_recent_failures(task_config.task_id)):
            priority_thresholds = sorted(task_config.legion.failure_priority_thresholds,
                                         key=lambda t: t.num_failures, reverse=True)
            for threshold in priority_thresholds:
                if len(recent_failures) >= threshold.num_failures:
                    return NotificationMsg(contents={'src': HOSTNAME,
                                                     'recent_output': recent_failures[0].output,
                                                     'recent_exit_code': recent_failures[0].exit_code},
                                           alert_key=f'[{HOSTNAME}][testudo][failure][{task_config.task_id}]',
                                           desc=f'Task [{task_config.task_id}] failed {len(recent_failures)} times',
                                           ttl=ceil(task_config.legion.reporting_delay_seconds * 2),
                                           priority=threshold.priority)
        if most_recent_run := first(tm.get_previous_results(task_config.task_id, limit=1)):
            return NotificationMsg(contents={'src': HOSTNAME,
                                             'recent_output': most_recent_run.output},
                                   desc=f'Task [{task_config.task_id} ran nominally',
                                   priority=Priority.INFO)
        return NotificationMsg(contents={'src': HOSTNAME},
                               alert_key=f'[{HOSTNAME}][testudo][unknown_task_id][{task_config.task_id}]',
                               desc=f'No runs detected for {task_config.task_id}',
                               ttl=ceil(task_config.legion.reporting_delay_seconds * 2),
                               priority=Priority.ERROR)

@typechecked
def publish_report(db: Path, task_config: TaskConfig) -> Optional[NotificationMsg]:
    rpt = report(db, task_config)
    try:
        legion_utils.broadcast_msg(task_config.legion.exchange, task_config.legion.route, rpt)
        return rpt
    except (AssertionError, SyntaxError):
        raise  # pragma: no cover # exists only for passing through errors in testing
    except Exception:  # pylint: disable=W0703
        log.error(f'Unable to report on task execution\nStack Trace:\n{format_exc()}')
        log.error(f'UNPUBLISHED REPORT:\n{pformat(rpt)}')
    return None


@typechecked
def run_reporter(db: Path, task_config: TaskConfig, halt: Event) -> None:
    while 42:
        publish_report(db, task_config)
        if halt.wait(task_config.legion.reporting_delay_seconds):
            return None

@typechecked
def start_reporter_daemon(db: Path, task_config: TaskConfig, halt: Optional[Event] = None) -> Thread:
    reporter_thread = Thread(target=run_reporter,
                             args=(db, task_config, halt),
                             daemon=True)
    reporter_thread.start()
    return reporter_thread



from pathlib import Path
from signal import signal, SIGINT
from threading import Event
from typing import Optional
from logging import DEBUG, INFO

import click
from typeguard import typechecked
from yaml import safe_load
from logzero import loglevel as set_loglevel

from testudo.config import TaskConfig
from testudo.log import log
from testudo.runner import run_with_reporter

DEFAULT_CONFIG = Path.cwd() / 'config.yaml'


@click.command()
@click.option('-c', '--config-file', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              writable=False, readable=True,
                              resolve_path=True, allow_dash=False),
              help="Path to the configuration file, defaults to ./config.yaml")
@click.option('-d', '--database-file', required=True,
              type=click.Path(file_okay=True, dir_okay=False,
                              writable=True, readable=True,
                              resolve_path=True, allow_dash=False),
              help="Path to the database file")
@click.option('--debug', is_flag=True, help='Set logging level to DEBUG')
@typechecked
def main(config_file: Optional[str], database_file: str, debug: bool) -> None:
    """A script for printing hello world based on some configured values"""
    set_loglevel(DEBUG if debug else INFO)
    halt_flag = Event()

    def handle_sigint(_, __):
        log.info("SIGINT Received, sending shutdown signal...")
        halt_flag.set()

    log.debug('Configuring SIGINT handler...')
    signal(SIGINT, handle_sigint)
    config_path = DEFAULT_CONFIG if config_file is None else Path(config_file)
    config = TaskConfig(**safe_load(config_path.open(encoding="utf8")))  # type: ignore
    run_with_reporter(Path(database_file), config, halt_flag)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['testudo']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'funcy>=1.16,<2.0',
 'legion-utils>=0.2.7,<0.3.0',
 'logzero>=1.7.0,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'typeguard>=2.13.2,<3.0.0',
 'types-pyyaml>=6.0.1,<7.0.0']

entry_points = \
{'console_scripts': ['testudo = testudo.cli:main']}

setup_kwargs = {
    'name': 'testudo',
    'version': '0.1.5',
    'description': 'A wrapper for commands to be run as periodic tasks while reporting on their results/errors to legiond. Intended to be integrated with supervisord, should work well with systemd in theory.',
    'long_description': '# Testudo\n\nA wrapper for commands to be run as periodic tasks while reporting on their results/errors to legiond. Intended to be integrated with supervisord, should work well with systemd in theory.\n\n## Setup & Usage\n\n### Installation\n\nTODO\n\n### Configuration\n\nTODO\n\n### Usage\n\nThe key reference for using `testudo` is:\n\n```bash\ntestudo --help\n```\n\n## Development\n\n### Standards\n\n- Be excellent to each other\n- Code coverage must be at 100% for all new code, or a good reason must be provided for why a given bit of code is not covered.\n  - Example of an acceptable reason: "There is a bug in the code coverage tool and it says its missing this, but its not".\n  - Example of unacceptable reason: "This is just exception handling, its too annoying to cover it".\n- The code must pass the following analytics tools. Similar exceptions are allowable as in rule 2.\n  - `pylint --disable=C0103,C0111,W1203,R0903,R0913 --max-line-length=120 testudo`\n  - `flake8 --max-line-length=120 ...`\n  - `mypy --ignore-missing-imports --follow-imports=skip --strict-optional ...`\n- All incoming information from users, clients, and configurations should be validated.\n- All internal arguments passing should be typechecked whenever possible with `typeguard.typechecked`\n\n### Development Setup\n\nUsing [poetry](https://python-poetry.org/) install from inside the repo directory:\n\n```bash\npoetry install\n```\n\nThis will set up a virtualenv which you can always activate with either `poetry shell` or run specific commands with `poetry run`. All instructions below that are meant to be run in the virtualenv will be prefaced with `(testudo)$ `\n\n#### IDE Setup\n\n**Sublime Text 3**\n\n```bash\ncurl -sSL https://gitlab.com/-/snippets/2066312/raw/master/poetry.sublime-project.py | poetry run python\n```\n\n#### Development\n\n### Testing\n\nAll testing should be done with `pytest` which is installed with the `dev` requirements.\n\nTo run all the unit tests, execute the following from the repo directory:\n\n```bash\n(testudo)$  pytest\n```\n\nThis should produce a coverage report in `/path/to/dewey-api/htmlcov/`\n\nWhile developing, you can use [`watchexec`](https://github.com/watchexec/watchexec) to monitor the file system for changes and re-run the tests:\n\n```bash\n(testudo)$ watchexec -r -e py,yaml pytest\n```\n\nTo run a specific test file:\n\n```bash\n(testudo)$ pytest tests/unit/test_cli.py\n```\n\nTo run a specific test:\n\n```bash\n(testudo)$ pytest tests/unit/test_cli.py::test_cli_basics\n```\n\nFor more information on testing, see the `pytest.ini` file as well as the [documentation](https://docs.pytest.org/en/stable/).\n',
    'author': 'Eugene Kovalev',
    'author_email': 'eugene@kovalev.systems',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/legion-robotnik/testudo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from collections import namedtuple
from contextlib import contextmanager
from itertools import takewhile
from pathlib import Path
from sqlite3 import Connection, connect
from time import time
from typing import Union, Optional, Generator
from uuid import UUID, uuid4 as uuid

from funcy import first
from typeguard import typechecked


TaskResult = namedtuple('TaskResult', ('result_id', 'task_id', 'timestamp',
                        'output', 'exit_code', 'stack_trace'))


@typechecked
class TaskManager:
    ALL_TABLES = """SELECT name FROM sqlite_schema
                    WHERE type IN ('table','view') 
                    AND name NOT LIKE 'sqlite_%'
                    ORDER BY 1;"""
    TASK_RESULTS_EXISTS = """SELECT name
                             FROM sqlite_master
                             WHERE type='table' AND name='task_results'"""
    CREATE_TASK_RESULTS = """CREATE TABLE task_results(
                             result_id CHARACTER(37) NOT NULL PRIMARY KEY,
                             task_id TEXT NOT NULL,
                             timestamp NUMERIC NOT NULL,
                             output TEXT NOT NULL,
                             exit_code INTEGER NOT NULL,
                             stack_trace TEXT)"""
    INSERT_TASK_RESULT = """INSERT INTO task_results VALUES (?, ?, ?, ?, ?, ?)"""
    GET_TASK_RESULT = """SELECT * FROM task_results WHERE result_id = ?"""
    GET_PREVIOUS_RESULTS = """SELECT * FROM task_results
                              WHERE task_id = ?
                              ORDER BY timestamp DESC
                              LIMIT ?"""
    GET_RESULTS_SORTED = """SELECT *
                            FROM task_results
                            WHERE task_id = ?
                            ORDER BY timestamp DESC"""

    def __init__(self, db: Path):
        self.db = str(db)
        self.connection: Optional[Connection] = None
        self._setup()

    @property
    def _connection(self) -> Connection:
        if self.connection is None:
            self.connection = connect(self.db)
        return self.connection

    def close(self) -> None:
        if self.connection is not None:
            self.connection.close()
            self.connection = None

    def _task_results_table_exists(self) -> bool:
        with self._connection as con:
            return any(row for row in con.execute(TaskManager.TASK_RESULTS_EXISTS))

    def _create_task_results_table(self) -> None:
        with self._connection as con:
            con.execute(TaskManager.CREATE_TASK_RESULTS)

    def _insert_task_result(self, result_id: Union[str, UUID],
                            task_id: str, timestamp: float,
                            output: str, exit_code: int,
                            stack_trace: Optional[str] = None) -> None:
        with self._connection as con:
            con.execute(TaskManager.INSERT_TASK_RESULT, (str(result_id), task_id,
                                                         timestamp, output,
                                                         exit_code, stack_trace))

    def insert_task_result(self, task_id: str, output: str, exit_code: int,
                           stack_trace: Optional[str] = None) -> Union[str, UUID]:
        timestamp = time()
        result_id = uuid()
        self._insert_task_result(result_id, task_id, timestamp,
                                 output, exit_code, stack_trace)
        return result_id

    def get_task_result(self, result_id: Union[str, UUID]) -> TaskResult:
        with self._connection as con:
            return TaskResult(*first(  # pragma: no branch
                row for row in con.execute(
                    TaskManager.GET_TASK_RESULT,
                    (str(result_id),))))

    def get_previous_results(self, task_id: str,
                             limit: int) -> Generator[TaskResult, None, None]:
        with self._connection as con:
            return (TaskResult(*row) for row in con.execute(
                TaskManager.GET_PREVIOUS_RESULTS, (task_id, limit)))

    def get_recent_failures(self, task_id: str) -> Generator[TaskResult, None, None]:
        with self._connection as con:
            cursor = con.cursor()
            cursor.execute(TaskManager.GET_RESULTS_SORTED, (task_id,))
            for result in takewhile(lambda r: r.exit_code != 0, (TaskResult(*row) for row in cursor)):
                yield result
            cursor.close()

    def _setup(self) -> None:
        if not self._task_results_table_exists():
            self._create_task_results_table()

    def report_success(self, task_id: str, result_output: str) -> str:
        return str(self.insert_task_result(task_id, result_output, 0))

    def report_failure(self, task_id: str, result_output: str,
                       exit_code: int, stack_trace: Optional[str] = None) -> str:
        return str(self.insert_task_result(task_id, result_output,
                                           exit_code, (stack_trace or '')))


@contextmanager
def task_manager(db: Path) -> Generator[TaskManager, None, None]:
    tm = TaskManager(db)
    try:
        yield tm
    finally:
        tm.close()
